package DataFlow;

import java.util.BitSet;

import AST.ASTAssignExpression;
import AST.ASTReferenceExpression;
import AST.ASTSymbol;
import AST.ASTVisitor;
import AST.SymbolTable;
import CFG.ASTCFGNode;
import CFG.CFG;

/** override bitset class so we can print it prettier */
@SuppressWarnings("serial")
class LivenessBitSet extends BitSet
{
	public LivenessBitSet(int size) {
		super(size);
	}
	
	@Override
	public String toString() {
		StringBuilder result = new StringBuilder();
		boolean first = true;
		for (int i = 0; i < size(); ++i) {
			if (!get(i))
				continue;
			
			if (!first) {
				result.append(" ");
			} else {
				first = false;
			}
			result.append(SymbolTable.getInstance().getSymbolWithId(i));
		}
		return result.toString();
	}
}

public class LivenessAnalysis implements DataFlowInterface<BitSet, ASTCFGNode> {
	private CFG<ASTCFGNode> cfg;
	private int maxId;

	@Override
	public void initialize(CFG<ASTCFGNode> cfg) {
		/* count number of variables:
		 *    in our simplistic scheme without declarations it's just the
		 *    number of known symbols */
		this.cfg = cfg;
		maxId = SymbolTable.getInstance().maxId();
	}

	/** Liveness is a backward analysis */
	@Override
	public boolean isBackward() {
		return true;
	}
	
	/** Visitor which create gen/kill Bitsets for a given ASTExpression */ 
	private static class CreateGenKillSets extends ASTVisitor.DefaultExpressionVisitor {
		private BitSet gen;
		private BitSet kill;
		
		public CreateGenKillSets(BitSet gen, BitSet kill) {
			this.gen = gen;
			this.kill = kill;
		}
		
		@Override
		public void visitAssignExpression(ASTAssignExpression assignExpression) {
			/* left side of assign is a kill */
			ASTSymbol var = assignExpression.getVariable();
			kill.set(var.getId());
			
			assignExpression.getExpression().accept(this);
		}

		@Override
		public void visitReferenceExpression(ASTReferenceExpression ref) {
			ASTSymbol variable = ref.getVariable();
			gen.set(variable.getId());
		}
	}
	
	public void createGenKill(ASTCFGNode node, BitSet gen, BitSet kill) {
		/* traverse AST to find ref/assign expression to create gen/kill sets */
		CreateGenKillSets visitor = new CreateGenKillSets(gen, kill);		
		node.accept(visitor);
	}

	@Override
	public BitSet transfer(ASTCFGNode node, BitSet data) {
		/* at the end of the program nothing is live, so the transfer function
		 * for the end block sets everything to false */
		if (node == cfg.getEnd()) {
			return new LivenessBitSet(maxId);
		}
		
		BitSet gen = new LivenessBitSet(maxId);
		BitSet kill = new LivenessBitSet(maxId);
		createGenKill(node, gen, kill);
		BitSet result = (BitSet) data.clone();
		result.andNot(kill);
		result.or(gen);
		return result;
	}

	/**
	 * We're interested in the least fixpoint. Perform union operation.
	 */
	@Override
	public BitSet union(BitSet d1, BitSet d2) {
		BitSet result = (BitSet) d1.clone();
		result.or(d2);
		return result;
	}
}
