package DataFlow;

import java.util.BitSet;

import AST.ASTAssignExpression;
import AST.ASTSymbol;
import AST.ASTVisitor;
import AST.SymbolTable;
import CFG.ASTCFG;
import CFG.ASTCFGNode;
import CFG.CFG;

/**
 * Reaching definition analysis using bitsets.
 * The tricky thing is that we have to map a pair (node, var) of CFGNode and variable to a single
 * bitnumber. This is done in the getIdFromNodeVarPair function.
 */
public class ReachingDefinitionAnalysis implements DataFlowInterface<BitSet, ASTCFGNode>{
	private ASTCFG cfg;
	private int maxId;
	
	/** transform a pair of a CFGNode and a variable to a unique id */
	public int getIdFromNodeVarPair(ASTCFGNode node, ASTSymbol variable) {
		int nodeId = (node == null) ? cfg.maxNodeId() : node.getId();
		return getIdFromNodeVarPair(nodeId, variable);
	}
	
	public int getIdFromNodeVarPair(int nodeId, ASTSymbol variable) {
		return nodeId * SymbolTable.getInstance().maxId() + variable.getId();
	}
	
	/** extract CFGNode from unique id */
	public ASTCFGNode blockFromId(int id) {
		int nodeid = id / SymbolTable.getInstance().maxId();
		/* the ? node */
		if (nodeid >= cfg.maxNodeId())
			return null;
		
		return cfg.nodeFromId(nodeid); 
	}
	
	/** extract symbol from unique id */
	public ASTSymbol symbolFromId(int id) {
		int symid = id % SymbolTable.getInstance().maxId();
		return SymbolTable.getInstance().getSymbolWithId(symid);
	}

	/**
	 * Initialize analysis: Mainly calculating the maximum id needed for
	 * the bitsets */
	@Override
	public void initialize(CFG<ASTCFGNode> ncfg) {
		this.cfg = (ASTCFG) ncfg;
		int maxSymbolId = SymbolTable.getInstance().maxId();
		int maxNodeId = cfg.maxNodeId(); /** we have an additional ? node */
		maxId = maxNodeId * maxSymbolId;
	}

	/**
	 * Reaching Definitions is a forward analysis
	 */
	@Override
	public boolean isBackward() {
		return false;
	}
	
	@Override
	public BitSet transfer(ASTCFGNode node, BitSet data) {
		/* create the (?, var) initial setting for the start node */
		if (node.equals(cfg.getStart())) {
			BitSet initial = new ReachingDefBitSet(this, maxId+1);
			for (int i = 0; i < SymbolTable.getInstance().maxId(); ++i) {
				ASTSymbol symbol = SymbolTable.getInstance().getSymbolWithId(i);
				int id = getIdFromNodeVarPair(null, symbol);
				initial.set(id);
			}
			return initial;
		}
		
		/********************************************
		 * TODO
		 * Bitte eigenen Code hier einbauen
		 ********************************************/
		return null;
	}

	/**
	 * We're interested in the least fixpoint. Perform a union of the bitsets.
	 */
	@Override
	public BitSet union(BitSet d1, BitSet d2) {
		/********************************************
		 * TODO
		 * Bitte eigenen Code hier einbauen
		 *******************************************/
		return null;
	}	
}

/**
 * Helper class which mainly override toString() to produce nicer output
 * for the dump
 */
@SuppressWarnings("serial")
class ReachingDefBitSet extends BitSet {
	public ReachingDefBitSet(ReachingDefinitionAnalysis analysis, int nBits) {
		super(nBits);
		this.analysis = analysis;
	}
	
	public String toString() {
		StringBuilder result = new StringBuilder();
		for (int i = 0; i < size(); ++i) {
			if (!get(i))
				continue;
			ASTSymbol symbol = analysis.symbolFromId(i);
			ASTCFGNode block = analysis.blockFromId(i);
			String id;
			if (block != null) {
				id = Integer.toString(block.getId());
			} else {
				id = "?";
			}
			result.append(String.format("(%s,%s) ", symbol, id));
		}
		return result.toString();
	}
	
	@Override
	public Object clone() {
		ReachingDefBitSet newBitset = new ReachingDefBitSet(analysis, size());
		newBitset.or(this);
		return newBitset;
	}
	
	private ReachingDefinitionAnalysis analysis;
}
